package dev.pfaff.glass.glfw;

import org.jspecify.annotations.NonNull;
import org.jspecify.annotations.Nullable;

import java.lang.foreign.Arena;
import java.lang.foreign.FunctionDescriptor;
import java.lang.foreign.Linker;
import java.lang.foreign.MemorySegment;
import java.lang.invoke.MethodHandle;
import java.nio.charset.StandardCharsets;
import java.util.Objects;

import static dev.pfaff.glass.glfw.GLFW.*;
import static java.lang.foreign.ValueLayout.*;

public record GLFWwindow(GLFW glfw, MemorySegment handle) {
	public GLFWwindow {
		Objects.requireNonNull(glfw, "glfw cannot be null");
		Objects.requireNonNull(handle, "handle cannot be null");
	}

	public static final MethodHandle MH_glfwMakeContextCurrent = LINKER.downcallHandle(FunctionDescriptor.ofVoid(ML_GLFWwindow));
	public static final MethodHandle MH_glfwGetCurrentContext = LINKER.downcallHandle(FunctionDescriptor.of(ML_GLFWwindow),
			Linker.Option.critical(false));
	public static final MethodHandle MH_glfwSwapInterval = LINKER.downcallHandle(FunctionDescriptor.ofVoid(JAVA_INT));
	public static final MethodHandle MH_glfwExtensionSupported = LINKER.downcallHandle(
			FunctionDescriptor.of(JAVA_BOOLEAN, ADDRESS.withTargetLayout(JAVA_BYTE)));
	public static final MethodHandle MH_glfwCreateWindow = LINKER.downcallHandle(
			FunctionDescriptor.of(ML_GLFWwindow,
					JAVA_INT,
					JAVA_INT,
					ADDRESS.withTargetLayout(JAVA_BYTE),
					ML_GLFWmonitor,
					ML_GLFWwindow));
	public static final MethodHandle MH_glfwWindowHint = LINKER.downcallHandle(FunctionDescriptor.ofVoid(JAVA_INT, JAVA_INT));
	public static final MethodHandle MH_glfwDestroyWindow = LINKER.downcallHandle(FunctionDescriptor.ofVoid(ML_GLFWwindow));
	public static final MethodHandle MH_glfwWindowShouldClose = LINKER.downcallHandle(FunctionDescriptor.of(JAVA_BOOLEAN, ML_GLFWwindow),
			Linker.Option.critical(false));
	public static final MethodHandle MH_glfwGetFramebufferSize = LINKER.downcallHandle(FunctionDescriptor.ofVoid(ML_GLFWwindow, ADDRESS.withTargetLayout(JAVA_INT), ADDRESS.withTargetLayout(JAVA_INT)),
			Linker.Option.critical(true));
	public static final MethodHandle MH_glfwGetWindowSize = LINKER.downcallHandle(FunctionDescriptor.ofVoid(ML_GLFWwindow, ADDRESS.withTargetLayout(JAVA_INT), ADDRESS.withTargetLayout(JAVA_INT)),
			Linker.Option.critical(true));
	public static final MethodHandle MH_glfwGetWindowContentScale = LINKER.downcallHandle(FunctionDescriptor.ofVoid(ML_GLFWwindow, ADDRESS.withTargetLayout(JAVA_FLOAT), ADDRESS.withTargetLayout(JAVA_FLOAT)),
			Linker.Option.critical(true));
	public static final MethodHandle MH_glfwSwapBuffers = LINKER.downcallHandle(FunctionDescriptor.ofVoid(ML_GLFWwindow));
	public static final MethodHandle MH_glfwShowWindow = LINKER.downcallHandle(FunctionDescriptor.ofVoid(ML_GLFWwindow));
	public static final MethodHandle MH_glfwHideWindow = LINKER.downcallHandle(FunctionDescriptor.ofVoid(ML_GLFWwindow));
	public static final MethodHandle MH_glfwSetWindowSize = LINKER.downcallHandle(FunctionDescriptor.ofVoid(ML_GLFWwindow, JAVA_INT, JAVA_INT));

	public static void setIntegerHint(GLFW glfw, WindowHint hint, int value) {
		try {
			MH_glfwWindowHint.invokeExact(glfw.MS_glfwWindowHint, hint.value(), value);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public static void setBooleanHint(GLFW glfw, WindowHint hint, boolean value) {
		setIntegerHint(glfw, hint, value ? 1 : 0);
	}

	@NonNull
	public static GLFWwindow create(GLFW glfw, int width, int height, String title, @Nullable GLFWmonitor monitor, @Nullable GLFWwindow share) {
		var monitorHandle = monitor != null ? monitor.handle() : MemorySegment.NULL;
		var shareHandle = share != null ? share.handle : MemorySegment.NULL;
		MemorySegment handle;
		try (var arena = Arena.ofConfined()) {
			var titlePtr = arena.allocateFrom(title, StandardCharsets.UTF_8);
			try {
				handle = (MemorySegment) MH_glfwCreateWindow.invokeExact(glfw.MS_glfwCreateWindow, width, height, titlePtr, monitorHandle, shareHandle);
			} catch (Throwable e) {
				throw new RuntimeException(e);
			}
		}
		if (handle.equals(MemorySegment.NULL)) throw new RuntimeException("Unable to create GLFW window");
		return new GLFWwindow(glfw, handle);
	}

	@Nullable
	public static GLFWwindow getCurrentContext(GLFW glfw) {
		MemorySegment handle;
		try {
			handle = (MemorySegment) MH_glfwGetCurrentContext.invokeExact(glfw.MS_glfwGetCurrentContext);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
		if (handle.equals(MemorySegment.NULL)) return null;
		return new GLFWwindow(glfw, handle);
	}

	public void makeContextCurrent() {
		try {
			MH_glfwMakeContextCurrent.invokeExact(glfw.MS_glfwMakeContextCurrent, handle);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public void destroy() {
		try {
			MH_glfwDestroyWindow.invokeExact(glfw.MS_glfwDestroyWindow, handle);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public boolean shouldClose() {
		try {
			return (boolean) MH_glfwWindowShouldClose.invokeExact(glfw.MS_glfwWindowShouldClose, handle);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public void setVisibility(boolean value) {
		try {
			if (value) {
				MH_glfwShowWindow.invokeExact(glfw.MS_glfwShowWindow, handle);
			} else {
				MH_glfwHideWindow.invokeExact(glfw.MS_glfwHideWindow, handle);
			}
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public record WindowSize(int width, int height) {}
	public record FramebufferSize(int width, int height) {}
	public record ContentScale(float x, float y) {
		public float min() {
			return Math.min(x, y);
		}

		public float max() {
			return Math.max(x, y);
		}
	}

	public WindowSize getWindowSize() {
		try {
			var slot = new long[1];
			var buf = MemorySegment.ofArray(slot);
			MH_glfwGetWindowSize.invokeExact(glfw.MS_glfwGetWindowSize, handle, buf, buf.asSlice(JAVA_INT.byteSize()));
			return new WindowSize(buf.getAtIndex(JAVA_INT, 0), buf.getAtIndex(JAVA_INT, 1));
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public FramebufferSize getFramebufferSize() {
		try {
			var slot = new long[1];
			var buf = MemorySegment.ofArray(slot);
			MH_glfwGetFramebufferSize.invokeExact(glfw.MS_glfwGetFramebufferSize, handle, buf, buf.asSlice(JAVA_INT.byteSize()));
			return new FramebufferSize(buf.getAtIndex(JAVA_INT, 0), buf.getAtIndex(JAVA_INT, 1));
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public ContentScale getWindowContentScale() {
		try {
			var slot = new long[1];
			var buf = MemorySegment.ofArray(slot);
			MH_glfwGetWindowContentScale.invokeExact(glfw.MS_glfwGetWindowContentScale, handle, buf, buf.asSlice(JAVA_FLOAT.byteSize()));
			return new ContentScale(buf.getAtIndex(JAVA_FLOAT, 0), buf.getAtIndex(JAVA_FLOAT, 1));
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public void swapBuffers() {
		try {
			MH_glfwSwapBuffers.invokeExact(glfw.MS_glfwSwapBuffers, handle);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public static void setSwapInterval(GLFW glfw, int interval) {
		try {
			MH_glfwSwapInterval.invokeExact(glfw.MS_glfwSwapInterval, interval);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public void setWindowSize(int width, int height) {
		try {
			MH_glfwSetWindowSize.invokeExact(glfw.MS_glfwSetWindowSize, handle, width, height);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * @return whether the named extension is supported by the current context.
	 */
	public static boolean isExtensionSupported(GLFW glfw, String name) {
		try (var arena = Arena.ofConfined()) {
			var namePtr = arena.allocateFrom(name, StandardCharsets.US_ASCII);
			try {
				return (boolean) MH_glfwExtensionSupported.invokeExact(glfw.MS_glfwExtensionSupported, namePtr);
			} catch (Throwable e) {
				throw new RuntimeException(e);
			}
		}
	}
}
