package dev.pfaff.glass.opengl;

import dev.pfaff.glass.OS;

import java.lang.foreign.*;
import java.nio.charset.StandardCharsets;
import java.util.Optional;

public final class GLSymbolLookup {
	public static final String LINUX_LIB_NAME = "libGL.so.1";
	public static final String MAC_OS_LIB_NAME = "/System/Library/Frameworks/OpenGL.framework/OpenGL";
	public static final String WINDOWS_LIB_NAME = "opengl32.dll";

	public static String libNameForOS(OS os) {
		return switch (os) {
			case OS.Linux() -> LINUX_LIB_NAME;
			// It is very important to use a string path, not wrapped in Path.of
			// This is because libraryLookup(Path, ...) will call path.toRealPath(), which will fail if the file does
			// not exist
			// The system frameworks, starting with recent versions of macOS, no longer exist on disk. Instead, dlopen
			// recognizes their paths and loads them from a system-wide "cache". Thus, the call to toRealPath will fail.
			case OS.macOS() -> MAC_OS_LIB_NAME;
			case OS.Windows _ -> WINDOWS_LIB_NAME;
			default -> throw new UnsupportedOperationException("Unsupported operating system: " + os);
		};
	}

	public static SymbolLookup symbolLookupForOS(OS os) {
		var lookupArena = Arena.global();
		var glLib = SymbolLookup.libraryLookup(libNameForOS(os), lookupArena);
		var linker = Linker.nativeLinker();
		return switch (OS.HOST) {
			case OS.Linux() -> {
				var MH_glXGetProcAddressARB = linker.downcallHandle(glLib.find("glXGetProcAddressARB").get(),
						FunctionDescriptor.of(ValueLayout.ADDRESS,
								ValueLayout.ADDRESS),
						Linker.Option.critical(true));
				yield name -> {
					assert !name.contains("\0") : "Invalid name";
					var nameBytes = (name + "\0").getBytes(StandardCharsets.US_ASCII);
					MemorySegment addr;
					try {
						addr = (MemorySegment) MH_glXGetProcAddressARB.invokeExact(MemorySegment.ofArray(nameBytes));
					} catch (Throwable e) {
						throw new RuntimeException(e);
					}
					if (addr.equals(MemorySegment.NULL)) {
						return glLib.find(name);
					}
					return Optional.of(addr);
				};
			}
			case OS.macOS() -> glLib;
			case OS.Windows _ -> {
				var MH_wglGetProcAddress = linker.downcallHandle(glLib.find("wglGetProcAddress").get(),
						FunctionDescriptor.of(ValueLayout.ADDRESS,
								ValueLayout.ADDRESS),
						Linker.Option.critical(true));
				yield name -> {
					assert !name.contains("\0") : "Invalid name";
					var nameBytes = (name + "\0").getBytes(StandardCharsets.US_ASCII);
					MemorySegment addr;
					try {
						addr = (MemorySegment) MH_wglGetProcAddress.invokeExact(MemorySegment.ofArray(nameBytes));
					} catch (Throwable e) {
						throw new RuntimeException(e);
					}
					if (addr.equals(MemorySegment.NULL)) {
						return glLib.find(name);
					}
					return Optional.of(addr);
				};
			}
			default -> throw new AssertionError();
		};
	}

	public static SymbolLookup symbolLookupForHost() {
		return symbolLookupForOS(OS.HOST);
	}

	private GLSymbolLookup() {}
}
