package dev.pfaff.glass.opengl;

import dev.pfaff.glass.opengl.GLParamTypeSets;
import org.jspecify.annotations.NonNull;
import org.jspecify.annotations.Nullable;

import java.lang.foreign.MemoryLayout;
import java.lang.foreign.MemorySegment;

import static java.lang.foreign.ValueLayout.*;

public sealed interface GLType {
	/**
	 * @return the class used in the public-facing Java API.
	 */
	@NonNull
	Class<?> javaClass();

	/**
	 * @return the class used in the native OpenGL API.
	 */
	@NonNull
	Class<?> nativeClass();

	@Nullable
	MemoryLayout layout();

	PointerInt POINTER_INT = new PointerInt();
	Void VOID = new Void();

	record Pointer(@Nullable GLType target, @Nullable Boolean mutable) implements GLType {
		@Override
		public Class<?> javaClass() {
			if (mutable == null) return MemorySegment.class;
			if (mutable) return MutPointer.class;
			return ConstPointer.class;
		}

		@Override
		public Class<?> nativeClass() {
			return MemorySegment.class;
		}

		@Override
		public MemoryLayout layout() {
			if (target != null) {
				MemoryLayout layout = target.layout();
				if (layout != null) {
					return ADDRESS.withTargetLayout(layout);
				}
			}
			return ADDRESS;
		}
	}

	final class PointerInt implements GLType {
		private PointerInt() {
		}

		@Override
		public Class<?> javaClass() {
			return long.class;
		}

		@Override
		public Class<?> nativeClass() {
			return ADDRESS.byteSize() == Integer.BYTES ? int.class : long.class;
		}

		@Override
		public MemoryLayout layout() {
			return ADDRESS.byteSize() == Integer.BYTES ? JAVA_INT : JAVA_LONG;
		}
	}

	final class Void implements GLType {
		private Void() {
		}

		@Override
		public Class<?> javaClass() {
			return void.class;
		}

		@Override
		public Class<?> nativeClass() {
			return void.class;
		}

		@Override
		public MemoryLayout layout() {
			return null;
		}
	}

	sealed interface Primitive extends GLType {
		@Override
		@NonNull
		Class<?> javaClass();

		@Override
		@NonNull
		default Class<?> nativeClass() {
			return javaClass();
		}

		record Boolean(MemoryLayout layout) implements Primitive {
			@Override
			public Class<?> javaClass() {
				return boolean.class;
			}
		}

		record Byte(MemoryLayout layout) implements Primitive {
			@Override
			public Class<?> javaClass() {
				return byte.class;
			}
		}

		record Short(MemoryLayout layout) implements Primitive {
			@Override
			public Class<?> javaClass() {
				return short.class;
			}
		}

		record Char(MemoryLayout layout) implements Primitive {
			@Override
			public Class<?> javaClass() {
				return char.class;
			}
		}

		record Int(MemoryLayout layout) implements Primitive {
			@Override
			public Class<?> javaClass() {
				return int.class;
			}
		}

		record Long(MemoryLayout layout) implements Primitive {
			@Override
			public Class<?> javaClass() {
				return long.class;
			}
		}

		record Float(MemoryLayout layout) implements Primitive {
			@Override
			public Class<?> javaClass() {
				return float.class;
			}
		}

		record Double(MemoryLayout layout) implements Primitive {
			@Override
			public Class<?> javaClass() {
				return double.class;
			}
		}
	}

	record Enum<T extends GLEnum>(Class<T> javaClass) implements GLType {
		@Override
		public Class<?> nativeClass() {
			return int.class;
		}

		@Override
		public MemoryLayout layout() {
			return GLParamTypeSets.ML_E();
		}
	}
}
