package dev.pfaff.glass.opengl;

import dev.pfaff.log4truth.NamedLogger;
import dev.pfaff.glass.opengl.enums.RawGLEnum;

import java.lang.constant.ConstantDescs;
import java.lang.foreign.*;
import java.lang.invoke.MethodHandle;
import java.lang.invoke.MethodHandles;
import java.lang.invoke.MethodType;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import static java.lang.invoke.MethodHandles.explicitCastArguments;
import static java.lang.invoke.MethodHandles.identity;
import static java.lang.invoke.MethodType.methodType;

final class Util {
	static final NamedLogger LOGGER = new NamedLogger("OpenGL");
	static final List<String> COMMAND_TAGS = List.of("GL_COMMAND");

	private static boolean getEnvFlag(String name) {
		return "1".equals(System.getenv(name));
	}
	static final boolean CHECKS = getEnvFlag("GL_CHECKS");
	static final boolean ERROR_CHECKS = getEnvFlag("GL_ERROR_CHECKS");
	static final boolean LOG_COMMANDS = getEnvFlag("GL_LOG_COMMANDS");
	static final Linker LINKER = Linker.nativeLinker();

	private static final Class<? extends FunctionDescriptor> C_FunctionDescriptorImpl;
	static final MethodHandle MH_FunctionDescriptorImpl_of;

	private static final MethodType MT_MethodHandle = methodType(MethodHandle.class);

	static {
		try {
			//noinspection unchecked
			C_FunctionDescriptorImpl = (Class) Class.forName("jdk.internal.foreign.FunctionDescriptorImpl");
			var M_FunctionDescriptorImpl_of = C_FunctionDescriptorImpl.getDeclaredMethod("of", MemoryLayout.class, List.class);
			M_FunctionDescriptorImpl_of.setAccessible(true);
			MH_FunctionDescriptorImpl_of = MethodHandles.lookup().unreflect(M_FunctionDescriptorImpl_of);
		} catch (ReflectiveOperationException e) {
			throw new RuntimeException(e);
		}
	}

	static final MethodHandle MH_newUnsupportedOperationException;
	private static final MethodHandle MH_GLenum_glEnum;
	private static final MethodHandle MH_cast_int_to_long = explicitCastArguments(identity(long.class), methodType(long.class, int.class));
	private static final MethodHandle MH_cast_long_to_int = explicitCastArguments(identity(int.class), methodType(int.class, long.class));
	private static final MethodHandle MH_RawGLEnum_fromGLEnum_as_GLEnum;
	private static final Map<Class<? extends GLEnum>, MethodHandle> MH_GLEnum_fromGLEnum = new ConcurrentHashMap<>();
	private static final MethodHandle MH_ConstPointer_newNullable;
	private static final MethodHandle MH_MutPointer_newNullable;
	private static final MethodHandle MH_ConstPointer_nullable_ms;
	private static final MethodHandle MH_MutPointer_nullable_ms;

	static final ValueLayout ADDRESS_SIZED_INT = ValueLayout.ADDRESS.byteSize() == 8
			? ValueLayout.JAVA_LONG
			: ValueLayout.JAVA_INT;
	static final Linker.Option[] EMPTY_LINKER_OPTIONS = new Linker.Option[0];

	static {
		var l = MethodHandles.lookup();
		try {
			MH_newUnsupportedOperationException = l.findConstructor(UnsupportedOperationException.class,
					methodType(void.class, String.class));
			MH_GLenum_glEnum = l.findVirtual(GLEnum.class, "glEnum", methodType(int.class));
			var MH_Pointer_nullable_ms = l.findStatic(Util.class, "pointerNullableMs", methodType(MemorySegment.class, Pointer.class));
			MH_ConstPointer_nullable_ms = MH_Pointer_nullable_ms.asType(MH_Pointer_nullable_ms.type().changeParameterType(0, ConstPointer.class));
			MH_MutPointer_nullable_ms = MH_Pointer_nullable_ms.asType(MH_Pointer_nullable_ms.type().changeParameterType(0, MutPointer.class));
			MH_RawGLEnum_fromGLEnum_as_GLEnum = l.findConstructor(RawGLEnum.class, methodType(void.class, int.class))
					.asType(methodType(GLEnum.class, int.class));
			MH_ConstPointer_newNullable = l.findStatic(ConstPointer.class, "newNullable", methodType(ConstPointer.class, MemorySegment.class));
			MH_MutPointer_newNullable = l.findStatic(MutPointer.class, "newNullable", methodType(MutPointer.class, MemorySegment.class));
		} catch (NoSuchMethodException | IllegalAccessException e) {
			throw new RuntimeException(e);
		}
	}

	private static MemorySegment pointerNullableMs(Pointer<?> pointer) {
		return pointer == null ? MemorySegment.NULL : pointer.ms();
	}

	static MethodHandle MH_toNType(Class<?> c, MemoryLayout layout) {
		if (c.isPrimitive()) {
			if (c == long.class) {
				if (layout.byteSize() == Integer.BYTES) {
					return MH_cast_long_to_int;
				} else {
					return null;
				}
			}
			return null;
		} else if (c == MemorySegment.class) {
			return null;
		} else if (GLEnum.class.isAssignableFrom(c)) {
			//noinspection unchecked
			return MH_GLEnum_to_native((Class<? extends GLEnum>) c);
		} else if (c == ConstPointer.class) {
			return MH_ConstPointer_nullable_ms;
		} else if (c == MutPointer.class) {
			return MH_MutPointer_nullable_ms;
		} else {
			throw new UnsupportedOperationException(STR."Unrecognized class for conversion to native type: \{c.getName()}");
		}
	}

	static MethodHandle MH_fromNType(Class<?> c, MemoryLayout layout) {
		if (c.isPrimitive()) {
			if (c == long.class) {
				if (layout.byteSize() == Integer.BYTES) {
					return MH_cast_int_to_long;
				} else {
					return null;
				}
			}
			return null;
		} else if (c == MemorySegment.class) {
			return null;
		} else if (GLEnum.class.isAssignableFrom(c)) {
			//noinspection unchecked
			return MH_GLEnum_from_native((Class<? extends GLEnum>) c);
		} else if (c == ConstPointer.class) {
			return MH_ConstPointer_newNullable;
		} else if (c == MutPointer.class) {
			return MH_MutPointer_newNullable;
		} else {
			throw new UnsupportedOperationException(STR."Unrecognized class for conversion from native type: \{c.getName()}");
		}
	}

	static MethodHandle MH_GLEnum_to_native(Class<? extends GLEnum> c) {
		return MH_GLenum_glEnum.asType(methodType(int.class, c));
	}

	static MethodHandle MH_GLEnum_from_native(Class<? extends GLEnum> c) {
		if (c == GLEnum.class) {
			return MH_RawGLEnum_fromGLEnum_as_GLEnum;
		} else {
			return MH_GLEnum_fromGLEnum.computeIfAbsent(c, clazz -> {
				try {
					return MethodHandles.lookup().findStatic(c, "fromGLEnum", methodType(clazz, int.class));
				} catch (NoSuchMethodException | IllegalAccessException e) {
					throw new RuntimeException(e);
				}
			});
		}
	}

	static GLType GLTypeFromClassAndLayout(Class<?> c, MemoryLayout layout) {
		if (c.isPrimitive()) {
			if (c == void.class) return GLType.VOID;
			if (c == boolean.class) return new GLType.Primitive.Boolean(layout);
			if (c == byte.class) return new GLType.Primitive.Byte(layout);
			if (c == short.class) return new GLType.Primitive.Short(layout);
			if (c == char.class) return new GLType.Primitive.Char(layout);
			if (c == int.class) return new GLType.Primitive.Int(layout);
			if (c == float.class) return new GLType.Primitive.Float(layout);
			if (c == double.class) return new GLType.Primitive.Double(layout);
			if (c == long.class) {
				if (layout.byteSize() == Integer.BYTES) {
					return GLType.POINTER_INT;
				}
				return new GLType.Primitive.Long(layout);
			}
			throw new AssertionError("Unhandled primitive class: " + c);
		} else if (c == MemorySegment.class) {
			if (!(layout instanceof AddressLayout)) {
				throw new IllegalArgumentException();
			}
			return new GLType.Pointer(null, null);
		} else if (GLEnum.class.isAssignableFrom(c)) {
			//noinspection unchecked
			return new GLType.Enum<>((Class<? extends GLEnum>) c);
		} else if (c == ConstPointer.class) {
			if (!(layout instanceof AddressLayout)) {
				throw new IllegalArgumentException();
			}
			return new GLType.Pointer(null, false);
		} else if (c == MutPointer.class) {
			if (!(layout instanceof AddressLayout)) {
				throw new IllegalArgumentException();
			}
			return new GLType.Pointer(null, true);
		} else {
			throw new UnsupportedOperationException(STR."Unrecognized class for conversion from native type: \{c.getName()}");
		}
	}

	static <T extends Throwable> RuntimeException throwUnchecked(Throwable throwable) throws T {
		throw (T) throwable;
	}

	static MethodHandle bindGLFunction(MethodHandles.Lookup lookup, String dynConstName, Class<?> type) {
		try {
			return UnlinkedGLFunction.link(MethodHandles.classData(lookup,
							ConstantDescs.DEFAULT_NAME,
							SymbolLookup.class),
					dynConstName,
					(MethodHandle) lookup.findStatic(UnboundGLFunctions.class,
							dynConstName,
							MT_MethodHandle).invokeExact());
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}
}
